#ifndef MIXSNET_COLOREDWEIGHTEDGRAPH_H
#define MIXSNET_COLOREDWEIGHTEDGRAPH_H
#include<map>
#include<vector>
#include<fstream>
#include<limits>
#include<string>
#include<sstream>
#include<stdlib.h>
#include<functional>
#include<sstream>
#include<boost/graph/adjacency_list.hpp>
#include<boost/numeric/ublas/matrix.hpp>
#include <boost/graph/johnson_all_pairs_shortest.hpp>

namespace mixsnet{

  typedef unsigned int color_type;
  const color_type color_default = 0;
  template<typename weight_type_t>
  class ColoredWeightedGraph : public boost::adjacency_list< boost::vecS, boost::vecS, boost::undirectedS,
  boost::property<boost::vertex_name_t, std::string, boost::property<boost::vertex_color_t, color_type> >,
  boost::property<boost::edge_weight_t, weight_type_t> >
  {
  protected:
    std::map<std::string, unsigned int> _str2index;
    weight_type_t _maxweight;
    void read(const std::string& filename){ // we suppose a single (i,j,w) or (j,i,w)
      std::ifstream edgestream;
      edgestream.open(filename.c_str());
      if (!edgestream.good()){
          std::cerr<<"Error: unable to open file "<<filename<<std::endl;
          exit(1);
      }
      _maxweight = std::numeric_limits<weight_type_t>::min();
      std::string line; getline(edgestream, line);
      while (line.size()) {
        std::istringstream linestream(line);
        std::string fromi; unsigned int i;
        linestream >> fromi;
        i = vertexIndexing(fromi);
        std::string toj; unsigned int j;
        linestream >> toj;
        j = vertexIndexing(toj);
        weight_type_t w;
        linestream >> w;
        if (i<j)
          std::swap(i,j);
        Edge enew =  boost::add_edge(i, j, *this).first;
        this->setWeight(enew, w);
        if (w>_maxweight)
          _maxweight = w;
        line.clear(); getline(edgestream, line);
      }
      edgestream.close();
    }
    unsigned int vertexIndexing(const std::string& name){
      unsigned int i;
      if (_str2index.find(name) != _str2index.end())
        i = _str2index[name];
      else{
          Vertex vnew = boost::add_vertex(*this);
          this->setName(vnew, name);
          this->setColor(vnew, color_default);
          i = this->getIndex(vnew);
          _str2index[name] = i;
      }
      return i;
    }
  public:
    typedef typename boost::graph_traits<ColoredWeightedGraph<weight_type_t> >::vertex_descriptor Vertex;
    typedef typename boost::graph_traits<ColoredWeightedGraph<weight_type_t> >::edge_descriptor Edge;

    ColoredWeightedGraph(){}
    ColoredWeightedGraph(const std::string& filename){
      read(filename);
    }
    virtual ~ColoredWeightedGraph(){}
    unsigned int getIndex(const Vertex& v) const{
      return get(boost::vertex_index, *this, v);
    }
    Vertex getVertexByIndex(unsigned int i) const{
      return vertex(i, *this);
    }
    unsigned int convertToIndex(const std::string& name){
      return _str2index[name];
    }
    std::string getName(const Vertex& v) const{
      return get(boost::vertex_name, *this, v);
    }
    void setName(const Vertex& v, const std::string& name){
      put(boost::vertex_name, *this, v, name);
    }
    std::string convertToName(unsigned int i){
      return getName(getVertexByIndex(i));
    }
    std::function<unsigned int (const std::string&)> getConvertToIndexFunction(){
      std::function<unsigned int (ColoredWeightedGraph<weight_type_t>*, const std::string&)> convertFunctor(&ColoredWeightedGraph<weight_type_t>::convertToIndex);
      std::function<unsigned int (const std::string&)> convertFunction = std::bind(convertFunctor, this, std::placeholders::_1);
      return convertFunction;
    }
    std::function<std::string (unsigned int)> getConvertToNameFunction(){
       std::function<std::string (ColoredWeightedGraph<weight_type_t>*, unsigned int)> convertFunctor(&ColoredWeightedGraph<weight_type_t>::convertToName);
       std::function<std::string (unsigned int)> convertFunction = std::bind(convertFunctor, this, std::placeholders::_1);
       return convertFunction;
    }
    color_type getColor(const Vertex& v) const{
      return get(boost::vertex_color, *this, v);
    }
    void setColor(const Vertex& v, color_type color){
      put(boost::vertex_color, *this, v, color);
    }
    void resetColor(){
      VertexIterator itv = verticesBegin(), itvend = verticesEnd();
      for(;itv != itvend; itv++)
        setColor(*itv, color_default);
    }
    const auto getColorProperty() const -> decltype(get(boost::vertex_color, *this)) { // Only >= g++4.7
      return get(boost::vertex_color, *this); // Return Type is a reference
      // (see doc for Boost Property Map Library)
    }
    weight_type_t getMaxWeight() const{
      return _maxweight;
    }
    weight_type_t getWeight(const Edge& e) const{
      return get(boost::edge_weight, *this, e);
    }
    void setWeight(const Edge& e, weight_type_t weight){
      put(boost::edge_weight, *this, e, weight);
    }
    unsigned int nbVertices() const{
      return num_vertices(*this);
    }
    typedef typename boost::graph_traits<ColoredWeightedGraph<weight_type_t> >::vertex_iterator VertexIterator;
    VertexIterator verticesBegin(){
      VertexIterator it;
      boost::tie(it, boost::tuples::ignore) =  vertices(*this);
      return it;
    }
    VertexIterator verticesEnd(){
      VertexIterator it;
      boost::tie(boost::tuples::ignore, it) = vertices(*this);
      return it;
    }
    typedef typename boost::graph_traits<ColoredWeightedGraph<weight_type_t> >::edge_iterator EdgeIterator;
    EdgeIterator edgesBegin(){
      EdgeIterator it;
      boost::tie(it, boost::tuples::ignore) = edges(*this);
      return it;
    }
    EdgeIterator edgesEnd(){
      EdgeIterator it;
      boost::tie(boost::tuples::ignore, it) = edges(*this);
      return it;
    }
    typedef typename boost::graph_traits<ColoredWeightedGraph<weight_type_t> >::out_edge_iterator OutEdgeIterator;
    OutEdgeIterator outEdgesBegin(const Vertex& v){
      OutEdgeIterator it;
      boost::tie(it, boost::tuples::ignore) = out_edges(v, *this);
      return it;
    }
    OutEdgeIterator outEdgesEnd(const Vertex& v){
      OutEdgeIterator it;
      boost::tie(boost::tuples::ignore, it) = out_edges(v, *this);
      return it;
    }
    template <typename weight_type_t1>
    friend std::ostream& operator<<(std::ostream& out, ColoredWeightedGraph<weight_type_t1>& g);
    template <typename weight_type_t1>
    friend std::istream& operator>>(std::istream& in, ColoredWeightedGraph<weight_type_t1>& g);
  };

  template <typename weight_type_t1>
  std::ostream& operator<<(std::ostream& out, ColoredWeightedGraph<weight_type_t1>& g){
    typename ColoredWeightedGraph<weight_type_t1>::VertexIterator itv = g.verticesBegin(), itvend = g.verticesEnd();
    out<<"# Colors\n";
    for(;itv != itvend; itv++)
      out<<g.getName(*itv)<<"\t"<<g.getColor(*itv)<<std::endl;
    return out;
  }

  template <typename weight_type_t1>
  std::istream& operator>>(std::istream& in, ColoredWeightedGraph<weight_type_t1>& g){
    std::string line;
    // # Colors
    getline(in,line); line.clear();
    unsigned int nbv = g.nbVertices();
    for (unsigned int i=0; i<nbv; i++){
        getline(in,line); std::istringstream linestream(line);
        std::string name; color_type c;
        linestream>>name; linestream>>c;
        g.setColor(vertex(g.convertToIndex(name), g), c);
        line.clear();
    }
    return in;
  }

  template<typename weight_type_t>
  class ColoredRoadNetwork : public ColoredWeightedGraph<weight_type_t>
  {
  protected:
    weight_type_t** _D; // operator [][] is required for boost shortest path
    void computeShortestPath(){
      unsigned int nbv = this->nbVertices();
      _D = new weight_type_t*[nbv];
      _D[0] = new weight_type_t[nbv*nbv];
      for (unsigned int i=1; i<nbv; i++) _D[i] = _D[i-1]+nbv;
      boost::johnson_all_pairs_shortest_paths(*this, _D);
//      for (unsigned int i=0; i<nbv; i++){
//          for (unsigned int j=0; j<nbv; j++){
//              //_Dinv[i][j]  = 1./_D[i][j];
//              std::cerr<<_D[i][j]<<" ";
//          }
//          std::cerr<<std::endl;
//      }
    }
  public:
    ColoredRoadNetwork()
      : ColoredWeightedGraph<weight_type_t>() {}
    ColoredRoadNetwork(const std::string& filename)
    : ColoredWeightedGraph<weight_type_t>(filename), _D(NULL) {}
    virtual ~ColoredRoadNetwork() {
      if (_D){
          delete[] _D[0]; delete[] _D;
      }
    }
    weight_type_t** allShortestPath(){
      if (!_D)
        computeShortestPath();
      return _D;
    }
    weight_type_t shortestPath(unsigned int i, unsigned int j){
      if (!_D)
        computeShortestPath();
      return _D[i][j];
    }
  };

  class EmptyColoredRoadNetwork : public ColoredRoadNetwork<double>
  {
  protected:
    void read(const std::string& filename){ // we read the first line with labels
      std::ifstream labelstream;
      labelstream.open(filename.c_str());
      if (!labelstream.good()){
          std::cerr<<"Error: unable to open file "<<filename<<std::endl;
          exit(1);
      }
      std::string line; getline(labelstream, line);
      std::istringstream linestream(line);
      std::string label;
      while(linestream >> label){
	vertexIndexing(label);
      }
      labelstream.close();
    }    
  public:  
    EmptyColoredRoadNetwork()
      : ColoredRoadNetwork<double>() {}
    EmptyColoredRoadNetwork(const std::string& labels)
      : ColoredRoadNetwork<double>() {
      read(labels);
    }
    virtual ~EmptyColoredRoadNetwork(){}
  };  
}
#endif
