#ifndef MIXSNET_INFLATEDDISTRIBUTION_H
#define MIXSNET_INFLATEDDISTRIBUTION_H

#include<limits>
#include<Functors.hpp>
#include<GaussianDistribution.hpp>

namespace mixsnet{

  template<class TGaussian>
  class InflatedDistribution : public TGaussian
  {
  protected:
    using TGaussian::_Nql;
    using TGaussian::_Muql;
    using TGaussian::_nbcol;
    using TGaussian::_dre;
    double _clump; // (clump)-inflated
    boost::numeric::ublas::matrix<unsigned int> _Nqli;
    boost::numeric::ublas::matrix<double> _Betaql;
    template<class TGaussian1>
    friend class UpdateNqlMuqlInflatedFunctor;
    template<class TGraph>
    void computeNqlMuqlNqliBetaqli(const TGraph& net, const Matrix<double>& Y);
    template<class TGaussian1>
    friend class UpdateSigma2qlInflatedFunctor;
    template<class TGraph>
    void computeSigma2ql(const TGraph& net, const Matrix<double>& Y);
    template<class TGaussian1>
    friend std::ostream& operator<<(std::ostream& out, InflatedDistribution<TGaussian>& dist);
    template<class TGaussian1>
    friend std::istream& operator>>(std::istream& in, InflatedDistribution<TGaussian>& dist);
  public:
    InflatedDistribution(color_type nbcol, double clump)
    : TGaussian(nbcol), _clump(clump),
      _Nqli(nbcol, nbcol, 0), _Betaql(nbcol, nbcol, 0.) {}
    virtual ~InflatedDistribution() {}
    void reinit(unsigned int nbcol) {*this = InflatedDistribution(nbcol, _clump);}
    double clump() const {return _clump;}
    inline double logdensity(unsigned int i, unsigned int j, color_type q, color_type l,  double x);
    double draw(unsigned int i, unsigned int j, color_type q, color_type l);
    template<class TGraph>
    void compute(const TGraph& net, const Matrix<double>& Y){
      computeNqlMuqlNqliBetaqli(net, Y);
      //std::cerr<<"_Nql "<<_Nql<<std::endl;
      //std::cerr<<"_Muql "<<_Muql<<std::endl;
      computeSigma2ql(net, Y);
      //std::cerr<<"_Sigma2ql "<<_Sigma2ql<<std::endl;
    }
    double getBetaql(color_type q, color_type l) const{
    	return _Betaql(q,l);
    }
    void setBetaql(color_type q, color_type l, double beta) {_Betaql(q,l) = beta;}
  };

  template<class TGaussian>
  double InflatedDistribution<TGaussian>::logdensity(unsigned int i, unsigned int j, color_type q, color_type l, double x)
  {
    if (q<l) std::swap(q, l);
    if (IsEqual()(_clump,x)){
        return log(_Betaql(q,l));
        //return 0;
    }
    else{
        return log(1-_Betaql(q,l)) + TGaussian::logdensity(i,j,q,l,x);
        //return TGaussian::logdensity(i,j,q,l,x);
    }
  }

  template<class TGaussian>
  double InflatedDistribution<TGaussian>::draw(unsigned int i, unsigned int j, color_type q, color_type l)
  {
    if (std::uniform_real_distribution<>()(_dre)<_Betaql(q,l))
      return _clump;
    else
      return TGaussian::draw(i,j,q,l);
  }

  template<class TGaussian>
  class UpdateNqlMuqlInflatedFunctor
  {
  private:
    InflatedDistribution<TGaussian> & _dist;
  public:
    UpdateNqlMuqlInflatedFunctor(HeteroscedasticGaussian& dist)
    :  _dist(static_cast<InflatedDistribution<TGaussian>& >(dist)) {}
    ~UpdateNqlMuqlInflatedFunctor() {}
    void operator() (unsigned int i, unsigned j, color_type ci, color_type cj, double yij){
      if (ci<cj) std::swap(ci, cj);
      if (IsEqual()(_dist._clump,yij))
        _dist._Nqli(ci, cj)++;
      else
        (UpdateNqlMuqlFunctor(_dist))(i,j,ci,cj,yij);
    }
    void exit() {}
  };

  template<class TGaussian>
  template<class TGraph>
  void InflatedDistribution<TGaussian>::computeNqlMuqlNqliBetaqli(const TGraph& net, const Matrix<double>& Y){
    _Nqli = boost::numeric::ublas::zero_matrix<unsigned int>(_nbcol, _nbcol);
    _Betaql = boost::numeric::ublas::zero_matrix<double>(_nbcol, _nbcol);
    TGaussian::template computeNqlMuql<TGraph, UpdateNqlMuqlInflatedFunctor<TGaussian> >(net, Y);
    for (color_type q = 0; q<_nbcol; q++){
        for (color_type l = 0; l<q; l++){
            unsigned int nql = _Nql(q,l);
            unsigned int nqli = _Nqli(q,l);
            if ((nql+nqli)>0){
                _Betaql(q,l) = double(nqli)/double(nql+nqli);
            }
        }
        unsigned int nqq = _Nql(q,q);
        unsigned int nqqi = _Nqli(q,q);
        if ((nqq+nqqi)>0)
          _Betaql(q,q) = double(nqqi)/double(nqq+nqqi);
    }
  }

  template<class TGaussian>
  class UpdateSigma2qlInflatedFunctor
  {
  private:
    InflatedDistribution<TGaussian>& _dist;
  public:
    UpdateSigma2qlInflatedFunctor(HeteroscedasticGaussian& dist)
      : _dist(static_cast<InflatedDistribution<TGaussian>& >(dist)) {}
    ~UpdateSigma2qlInflatedFunctor() {}
    void operator() (unsigned int i, unsigned j, color_type ci, color_type cj, double yij){
      if (!IsEqual()(_dist._clump,yij)){
          (UpdateSigma2qlFunctor(_dist))(i,j,ci,cj,yij);
      }
    }
    void exit() {}
  };

  template<class TGaussian>
  template<class TGraph>
  void InflatedDistribution<TGaussian>::computeSigma2ql(const TGraph& net, const Matrix<double>& Y)
  {
    TGaussian::template computeSigma2ql<TGraph, UpdateSigma2qlInflatedFunctor<TGaussian> >(net, Y);
  }

  template<class TGaussian>
  std::ostream& operator<<(std::ostream& out, InflatedDistribution<TGaussian>& dist)
  {
    out<<static_cast<TGaussian&>(dist);
    out<<"# Beta\n";
    for (color_type q = 0; q<dist.nbColors(); q++){
      for (color_type l = 0; l<=q; l++)
        out<<dist.getBetaql(q,l)<<" ";
      out<<std::endl;
    }
    return out;
  }

  template<class TGaussian>
  std::istream& operator>>(std::istream& in, InflatedDistribution<TGaussian>& dist)
  {
	in>>static_cast<TGaussian&>(dist);
    std::string line;
    double x;
    // # Beta
    getline(in,line);
    for (color_type q = 0; q<dist.nbColors(); q++){
        getline(in,line); std::istringstream linestream(line);
        for (color_type l = 0; l<=q; l++){
          linestream>>x;
          dist.setBetaql(q,l,x);
        }
        line.clear();
    }
    //std::cerr<<"_Nql "<<dist._Nql<<std::endl;
    //std::cerr<<"_Muql "<<dist._Muql<<std::endl;
    //std::cerr<<"_Sigma2ql "<<dist._Sigma2ql<<std::endl;
    //std::cerr<<"_Betaql "<<dist._Sigma2ql<<std::endl;
    return in;
  }

}
#endif
